#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
#                                       (@@@@%%%%%%%%%&@@&.
#                              /%&&%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%&@@(
#                              *@&%%%%%%%%%&&%%%%%%%%%%%%%%%%%%&&&%%%%%%%
#                               @   @@@(@@@@%%%%%%%%%%%%%%%%&@@&* @@@   .
#                               ,   .        .  .@@@&                   /
#                                .       .                              *
#                               @@              .                       @
#                              @&&&&&&@. .    .                     *@%&@
#                              &&&&&&&&&&&&&&&&@@        *@@############@
#                     *&/ @@ #&&&&&&&&&&&&&&&&&&&&@  ###################*
#                              @&&&&&&&&&&&&&&&&&&##################@
#                                 %@&&&&&&&&&&&&&&################@
#                                        @&&&&&&&&&&%#######&@%
#  nanaimo                                   (@&&&&####@@*
#
import pathlib
import typing

import pytest

import nanaimo
import nanaimo.fixtures
import nanaimo.pytest.plugin


class Fixture(nanaimo.fixtures.SubprocessFixture):
    """
    This fixture assumes that scp is available and functional on the system.
    """

    fixture_name = 'nanaimo_scp'
    argument_prefix = 'scp'

    @classmethod
    def on_visit_test_arguments(cls, arguments: nanaimo.Arguments) -> None:
        super().on_visit_test_arguments(arguments)
        arguments.add_argument('--file', help='The file to upload.')
        arguments.add_argument('--remote-dir', help='The directory to upload to.')
        arguments.add_argument('--target',
                               help='The IP or hostname for the target system.')
        arguments.add_argument('--as-user',
                               help='The user to upload as.')
        arguments.add_argument('--identity',
                               help='The identify file to use')

    def on_construct_command(cls, args: nanaimo.Namespace, inout_artifacts: nanaimo.Artifacts) -> str:
        """
        Form the upload command.
        """
        scp_file = cls.get_arg_covariant_or_fail(args, 'file')
        scp_port = cls.get_arg_covariant(args, 'port')
        scp_identity = cls.get_arg_covariant(args, 'identity')
        scp_as_user = cls.get_arg_covariant(args, 'as_user')
        scp_target = cls.get_arg_covariant_or_fail(args, 'target')

        remote_directory = pathlib.Path(cls.get_arg_covariant_or_fail(args, 'remote_dir'))
        remote_path = remote_directory / pathlib.Path(scp_file).name
        port_string = '-P {}'.format(scp_port) if scp_port is not None else ''
        identity_string = '-i {}'.format(scp_identity) if scp_identity is not None else ''
        scp_as_user_string = ('{}@'.format(scp_as_user) if scp_as_user is not None else scp_as_user)

        setattr(inout_artifacts, 'remote_path', remote_path)

        cmd = 'scp {ident} {port}{file} {user}{target}:{remote_path}'.format(port=port_string,
                                                                             file=str(scp_file),
                                                                             user=scp_as_user_string,
                                                                             target=scp_target,
                                                                             remote_path=str(remote_path),
                                                                             ident=identity_string)
        return cmd


@nanaimo.fixtures.PluggyFixtureManager.type_factory
def get_fixture_type() -> typing.Type['Fixture']:
    return Fixture


@pytest.fixture
def nanaimo_scp(request: typing.Any) -> nanaimo.fixtures.Fixture:
    """
    The SCP fixture allows a test to use SSH to copy files to a target system.
    This `pytest fixture <https://docs.pytest.org/en/latest/fixture.html>`_ provides
    a :class:`nanaimo.builtin.nanaimo_scp.Fixture` fixture to a pytest.

    :param pytest_request: The request object passed into the pytest fixture factory.
    :type pytest_request: _pytest.fixtures.FixtureRequest
    :rtype: nanaimo.builtin.nanaimo_scp.Fixture
    """
    return nanaimo.pytest.plugin.create_pytest_fixture(request, Fixture.get_canonical_name())



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
#                                       (@@@@%%%%%%%%%&@@&.
#                              /%&&%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%&@@(
#                              *@&%%%%%%%%%&&%%%%%%%%%%%%%%%%%%&&&%%%%%%%
#                               @   @@@(@@@@%%%%%%%%%%%%%%%%&@@&* @@@   .
#                               ,   .        .  .@@@&                   /
#                                .       .                              *
#                               @@              .                       @
#                              @&&&&&&@. .    .                     *@%&@
#                              &&&&&&&&&&&&&&&&@@        *@@############@
#                     *&/ @@ #&&&&&&&&&&&&&&&&&&&&@  ###################*
#                              @&&&&&&&&&&&&&&&&&&##################@
#                                 %@&&&&&&&&&&&&&&################@
#                                        @&&&&&&&&&&%#######&@%
#  nanaimo                                   (@&&&&####@@*
#
import asyncio
import typing

import pytest

import nanaimo
import nanaimo.fixtures
import nanaimo.pytest.plugin


class Fixture(nanaimo.fixtures.Fixture):
    """
    This fixture takes a list of other fixtures and runs them concurrently returning a
    :meth:`nanaimo.Artifacts.combine` ed set of :class:`nanaimo.Artifacts`.

    You can use this fixture directly in your unit tests:

    .. invisible-code-block: python

        from nanaimo.fixtures import FixtureManager
        from nanaimo import Namespace, Arguments, Artifacts
        from nanaimo.builtin import nanaimo_bar, nanaimo_gather
        import asyncio
        import argparse

        loop = asyncio.get_event_loop()
        manager = FixtureManager(loop=loop)

    .. code-block:: python

        async def example1() -> Artifacts:

            bar_one = nanaimo_bar.Fixture(manager)
            bar_two = nanaimo_bar.Fixture(manager)
            gather_fixture = nanaimo_gather.Fixture(manager)

            return await gather_fixture.gather(
                gather_coroutine=[
                    bar_one.gather(bar_number=1),
                    bar_two.gather(bar_number=2)
                ]
            )

    .. invisible-code-block: python

        result = loop.run_until_complete(example1())

        assert 'bar_1' in result
        assert 'bar_2' in result

    You can also use the --gather-coroutines argument to specify fixtures by name::

        nait nanaimo_gather --gather-coroutine nanaimo_bar --gather-coroutine nanaimo_bar

    """

    fixture_name = 'nanaimo_gather'
    argument_prefix = 'gather'

    @classmethod
    def on_visit_test_arguments(cls, arguments: nanaimo.Arguments) -> None:
        arguments.add_argument('--coroutine', action='append')

    async def on_gather(self, args: nanaimo.Namespace) -> nanaimo.Artifacts:
        """
        Multi-plex fixture
        """
        fixture_list = []  # type: typing.List[typing.Coroutine]
        gather_coroutine = self.get_arg_covariant_or_fail(args, 'coroutine')

        for arg in gather_coroutine:
            if asyncio.iscoroutine(arg):
                fixture_list.append(arg)
            else:
                fixture_list.append(self.manager.create_fixture(arg, args).gather())

        results = await asyncio.gather(*fixture_list, loop=self.loop)
        return nanaimo.Artifacts.combine(*results)


@nanaimo.fixtures.PluggyFixtureManager.type_factory
def get_fixture_type() -> typing.Type['Fixture']:
    return Fixture


@pytest.fixture
def nanaimo_gather(request: typing.Any) -> nanaimo.fixtures.Fixture:
    """
    Provides a :class:`nanaimo.builtin.nanaimo_gather.Fixture` fixture to a pytest.

    :param pytest_request: The request object passed into the pytest fixture factory.
    :type pytest_request: _pytest.fixtures.FixtureRequest
    :rtype: nanaimo.builtin.nanaimo_gather.Fixture
    """
    return nanaimo.pytest.plugin.create_pytest_fixture(request, Fixture.get_canonical_name())



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
#                                       (@@@@%%%%%%%%%&@@&.
#                              /%&&%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%&@@(
#                              *@&%%%%%%%%%&&%%%%%%%%%%%%%%%%%%&&&%%%%%%%
#                               @   @@@(@@@@%%%%%%%%%%%%%%%%&@@&* @@@   .
#                               ,   .        .  .@@@&                   /
#                                .       .                              *
#                               @@              .                       @
#                              @&&&&&&@. .    .                     *@%&@
#                              &&&&&&&&&&&&&&&&@@        *@@############@
#                     *&/ @@ #&&&&&&&&&&&&&&&&&&&&@  ###################*
#                              @&&&&&&&&&&&&&&&&&&##################@
#                                 %@&&&&&&&&&&&&&&################@
#                                        @&&&&&&&&&&%#######&@%
#  nanaimo                                   (@&&&&####@@*
#
import socket
import textwrap
import typing

import pytest
import serial

import nanaimo
import nanaimo.config
import nanaimo.connections
import nanaimo.fixtures
import nanaimo.pytest.plugin
import nanaimo.version


class _CharacterDisplayInterface:

    def __init__(self) -> None:
        pass

    def write(self, input_line: typing.Union[typing.List[str], str]) -> None:
        """
        Write text to the display.
        """
        pass

    def clear(self, display_default_message: bool = True) -> None:
        """
        Clear the display.
        """
        pass

    def set_default_message(self, message: str) -> None:
        """
        Set the message to display when the screen is cleared.
        """
        pass

    def set_bg_colour(self, r: int, g: int, b: int) -> None:
        pass

    def configure(self) -> None:
        """
        Configure the display using the arguments provided to this object.
        """
        pass


class _CharacterDisplayAdafruitSerialBackpack(_CharacterDisplayInterface):
    """
    A character display that uses the
    'Adafruit Serial Backpack command protocol \
    <https://learn.adafruit.com/usb-plus-serial-backpack/command-reference>'_

    ``nanaimo.cfg`` configuration ::

        [character_display]
        serial_port=/dev/your_devices_serial_port
        default_message=This is the default message.
        columns=20
        lines=4

    """

    @classmethod
    def on_visit_test_arguments(cls, arguments: nanaimo.Arguments) -> None:
        arguments.add_argument('--serial-port', help='The serial port the character display is attached to.')

    def __init__(self, args: nanaimo.Namespace) -> None:
        port = args.character_display_serial_port
        self._uart = serial.Serial(port)
        self.display_on()
        self._columns = int(args.character_display_columns)
        self._lines = int(args.character_display_lines)
        self._wrapper = textwrap.TextWrapper(width=self._columns, drop_whitespace=True)
        # Enable auto-scroll
        self._uart.write(b'\xFE\x51')
        self._builtin_values = {
            'version': nanaimo.version.__version__,
            'nl': '\n',
            'hostname': socket.gethostname()
        }
        self._default_message = None  # type: typing.Optional[typing.List[str]]
        default_message = args.character_display_default_message
        if default_message is not None:
            self._default_message = default_message.format(**self._builtin_values).split('\n')

    # +-----------------------------------------------------------------------+
    # | CharacterDisplay
    # +-----------------------------------------------------------------------+
    def write(self, input_line: typing.Union[typing.List[str], str]) -> None:
        if not isinstance(input_line, list):
            lines = self._wrapper.wrap(input_line)
        else:
            lines = input_line

        for line_number in range(0, len(lines)):
            self._uart.write(bytes([0xFE, 0x47, 0x01, line_number + 1]))
            self._uart.write(lines[line_number].encode('ascii'))

    def clear(self, display_default_message: bool = True) -> None:
        self._uart.write(b'\xFE\x58')
        if display_default_message and self._default_message is not None:
            self.write(self._default_message)

    # +-----------------------------------------------------------------------+
    # | ADAFRUIT USB+SERIAL LCD BACKPACK
    # |   https://www.adafruit.com/product/781
    # +-----------------------------------------------------------------------+
    def go_home(self) -> None:
        self._uart.write(b'\xFE\x48')

    def display_on(self) -> None:
        self._uart.write(b'\xFE\x42')

    def display_off(self) -> None:
        self._uart.write(b'\xFE\x46')

    def set_bg_colour(self, r: int, g: int, b: int) -> None:
        self._uart.write(bytes([0xFE, 0xD0, r, g, b]))

    def configure(self) -> None:
        self._uart.write(bytes([0xFE, 0xD1, self._columns, self._lines]))
        if self._default_message is not None:
            self._uart.write(bytes([0xFE, 0x40]))
            for line in self._default_message:
                line_len = len(line)
                self._uart.write(line.encode('ascii'))
                if line_len < self._columns:
                    self._uart.write((' ' * (self._columns - line_len)).encode('ascii'))


class CharacterDisplay(nanaimo.fixtures.Fixture):
    """
    A null display that can be used as a no-op when no display is available and
    is also the common base of all character display objects in Nanimo.

    ``nanaimo.cfg`` configuration ::

        [character_display]
        default_message=This will be displayed when clear is called.

    """

    fixture_name = 'character_display'

    @classmethod
    def _create_display(cls, args: typing.Optional[nanaimo.Namespace]) -> _CharacterDisplayInterface:
        if args is None:
            defaults = nanaimo.config.ArgumentDefaults.create_defaults_with_early_rc_config()
            args = nanaimo.Namespace(defaults=defaults)
        try:
            return _CharacterDisplayAdafruitSerialBackpack(args)
        except serial.SerialException:
            pass
        # No hardware display. Just use the null display.
        return _CharacterDisplayInterface()

    @classmethod
    def on_visit_test_arguments(cls, arguments: nanaimo.Arguments) -> None:
        _CharacterDisplayAdafruitSerialBackpack.on_visit_test_arguments(arguments)
        arguments.add_argument('--default-message', help='A message to display when the screen is cleared.')
        arguments.add_argument('--columns', default=16, help='The number of characters in a line of text for '
                               'this display.')
        arguments.add_argument('--lines', default=2, help='The number of lines on the display.')
        arguments.add_argument('--write', help='A string to write to the display.')
        arguments.add_argument('--configure', action='store_true', help='Configures the display to use '
                               'configured values (e.g. columns and lines).')
        arguments.add_argument('--clear', action='store_true', help='Clear the screen.')

    def __init__(self,
                 manager: 'nanaimo.fixtures.FixtureManager',
                 args: typing.Optional[nanaimo.Namespace] = None,
                 **kwargs: typing.Any):
        super().__init__(manager, args, **kwargs)
        self._impl = self._create_display(args)

    def write(self, input_line: str) -> None:
        """
        Write text to the display.
        """
        try:
            self._impl.write(input_line)
        except serial.SerialException:
            pass

    def clear(self, display_default_message: bool = True) -> None:
        """
        Clear the display.
        """
        try:
            self._impl.clear(display_default_message)
        except serial.SerialException:
            pass

    def configure(self) -> None:
        """
        Configure the display using the arguments provided to this object.
        """
        try:
            self._impl.configure()
        except serial.SerialException:
            pass

    def set_bg_colour(self, r: int, g: int, b: int) -> None:
        try:
            self._impl.set_bg_colour(r, g, b)
        except serial.SerialException:
            pass

    async def on_gather(self, args: nanaimo.Namespace) -> nanaimo.Artifacts:
        if args.character_display_configure:
            self._impl.configure()

        if args.character_display_clear:
            self._impl.clear()

        if args.character_display_write is not None:
            self._impl.write(args.character_display_write)

        return nanaimo.Artifacts()


@nanaimo.fixtures.PluggyFixtureManager.type_factory
def get_fixture_type() -> typing.Type['nanaimo.fixtures.Fixture']:
    return CharacterDisplay


@pytest.fixture
def nanaimo_character_display(request: typing.Any) -> nanaimo.fixtures.Fixture:
    """
    Provides a :class:`nanaimo.display.CharacterDisplay` fixture to pytest. See
    `this Adafruit product <https://www.adafruit.com/product/782>`_ for an example of the display
    type this fixture supports.

    :param pytest_request: The request object passed into the pytest fixture factory.
    :type pytest_request: _pytest.fixtures.FixtureRequest
    :return: A fixture providing access to an attached character-oriented display.
    :rtype: nanaimo.display.CharacterDisplay
    """
    return nanaimo.pytest.plugin.create_pytest_fixture(request, CharacterDisplay.get_canonical_name())



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
#                                       (@@@@%%%%%%%%%&@@&.
#                              /%&&%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%&@@(
#                              *@&%%%%%%%%%&&%%%%%%%%%%%%%%%%%%&&&%%%%%%%
#                               @   @@@(@@@@%%%%%%%%%%%%%%%%&@@&* @@@   .
#                               ,   .        .  .@@@&                   /
#                                .       .                              *
#                               @@              .                       @
#                              @&&&&&&@. .    .                     *@%&@
#                              &&&&&&&&&&&&&&&&@@        *@@############@
#                     *&/ @@ #&&&&&&&&&&&&&&&&&&&&@  ###################*
#                              @&&&&&&&&&&&&&&&&&&##################@
#                                 %@&&&&&&&&&&&&&&################@
#                                        @&&&&&&&&&&%#######&@%
#  nanaimo                                   (@&&&&####@@*
#
"""
TODO: See https://github.com/ppannuto/python-saleae/blob/master/saleae/saleae.py for the command
strings. They don't seem to be documented anywhere else.
"""

import asyncio
import typing

import pytest

import nanaimo
import nanaimo.fixtures
import nanaimo.pytest.plugin


class Fixture(nanaimo.fixtures.Fixture):
    """
    .. Warning:: Stubbed-out implementation
        This fixture doesn't do anything yet either then the most naive query possible of the Saleae.
        This would be a great first contribution to the Nanaimo project.
    """

    fixture_name = 'nanaimo_saleae'
    argument_prefix = 'sae'

    @classmethod
    def on_visit_test_arguments(cls, arguments: nanaimo.Arguments) -> None:
        arguments.add_argument('--port',
                               enable_default_from_environ=True,
                               default='10429',
                               help='TCP port for the logic socket server.')
        arguments.add_argument('--host',
                               enable_default_from_environ=True,
                               default='localhost',
                               help='hostname for the logic socket server.')

    async def on_gather(self, args: nanaimo.Namespace) -> nanaimo.Artifacts:
        saleae_host = self.get_arg_covariant_or_fail(args, 'saleae_host')
        saleae_port = self.get_arg_covariant_or_fail(args, 'saleae_port')

        self.logger.info('about to connect to {}:{}'.format(saleae_host, saleae_port))
        reader, writer = await asyncio.open_connection(host=saleae_host, port=saleae_port, loop=self.loop)

        writer.write('GET_ALL_SAMPLE_RATES\0'.encode('utf-8'))

        rx_buffer = ''
        while 'ACK' not in rx_buffer:
            rx_bytes = await reader.read(256)
            if rx_bytes is not None:
                rx_text = rx_bytes.decode('utf-8')
                self.logger.debug('rx: %s', rx_text)
                rx_buffer += rx_text
                # this is kind of crap. do way better.
                if 'NAK' == rx_buffer[0:3]:
                    self.logger.error('NAK')
                    break

        writer.close()
        await writer.wait_closed()
        return nanaimo.Artifacts()


@nanaimo.fixtures.PluggyFixtureManager.type_factory
def get_fixture_type() -> typing.Type['Fixture']:
    return Fixture


@pytest.fixture
def nanaimo_instr_saleae(request: typing.Any) -> nanaimo.fixtures.Fixture:
    """
    Provides a :class:`nanaimo.instruments.saleae.Fixture` fixture to a pytest.
    This fixture controls a `Saleae logic analyser <https://www.saleae.com/>`_
    attached to the system via USB.

    :param pytest_request: The request object passed into the pytest fixture factory.
    :type pytest_request: _pytest.fixtures.FixtureRequest
    :return: A fixture providing control of a Saleae logic analyser.
    :rtype: nanaimo.instruments.saleae.Fixture
    """
    return nanaimo.pytest.plugin.create_pytest_fixture(request, Fixture.get_canonical_name())



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
"""
Also see the YPR class's doctests for nominal case coverage.
"""
import pathlib
import typing

from nanaimo.parsers.imu import YPR


class Handler:
    def __init__(self) -> None:
        self.records = []  # type: typing.List[typing.Tuple[float, float, float]]

    def __call__(self, record: typing.Tuple[float, float, float]) -> None:
        self.records.append(record)


def common_test(binary_data: typing.List[bytes]) -> typing.Tuple[Handler, YPR]:

    handler = Handler()

    parser = YPR(handler)

    for chunk in binary_data:
        parser.push_bytes(chunk)

    return (handler, parser)


def test_line_too_long() -> None:
    '''
    Ensure we discard data if it is too long. This ensures we don't fill available memory up
    with junk data.
    '''
    binary_data = [
        ('a' * (YPR.max_line_length + 1)).encode('ascii'),
        '\n'.encode('ascii')
    ]

    handler, parser = common_test(binary_data)
    assert parser.bytes_discarded == YPR.max_line_length + 1


def test_junk_data() -> None:
    """
    Make sure we continue to parse after encountering unknown data.
    """
    binary_data = [(
        '#YPR=104.65,-3.73,96.25\n' +
        '#YPR=103.87,-3.60,96.24\n' +
        '#YPR=102.71,-3.45,99.21\n' +
        '#YPR=101.29,-3.39,99.14\n' +
        '#YPR=this is a malformed field\n'
        '#YPR=97.29,-2.89,99.08\n' +
        '#YPR=97.26,-3.72,97.16\n' +
        '#YPR=97.99,-3.98,96.61\n'
    ).encode('ascii')]

    handler, parser = common_test(binary_data)
    assert parser.line_errors == 1
    assert len(handler.records) == 7


def test_integers() -> None:
    """
    We do allow integers in our data.
    """
    binary_data = [(
        '#YPR=1,2,3\n'
    ).encode('ascii')]

    handler, parser = common_test(binary_data)
    assert parser.line_errors == 0
    assert len(handler.records) == 1
    assert handler.records[0][0] == 1
    assert handler.records[0][1] == 2
    assert handler.records[0][2] == 3


def test_bulk_data(paths_for_test: typing.Any) -> None:
    """
    Test handling a larger amount of data.
    """
    handler = Handler()

    parser = YPR(handler)

    with open(str(paths_for_test.imu_text), 'r') as textual_data:
        for line in textual_data:
            parser.push_bytes(line.encode('ascii'))

    assert len(handler.records) == 718

    try:
        import numpy as np
        import matplotlib.pyplot as plt

        ypr_data = np.array(handler.records).astype(np.float)
        Y, P, R = ypr_data[:, 0], ypr_data[:, 1], ypr_data[:, 2]

        fig = plt.figure()
        ax = fig.add_subplot(111, projection='rectilinear')
        ax.scatter(Y, range(0, len(Y)), c='brown', s=1, alpha=0.5).set_label('yaw')
        ax = fig.add_subplot(111, projection='rectilinear')
        ax.scatter(P, range(0, len(P)), c='orange', s=1, alpha=0.5).set_label('pitch')
        ax = fig.add_subplot(111, projection='rectilinear')
        ax.scatter(R, range(0, len(R)), c='blue', s=1, alpha=0.5).set_label('roll')

        ax.set_xlabel('deg.')
        ax.set_ylabel('sample')
        legend = ax.legend(loc="lower left", fontsize=8)

        # Thanks stackoverflow!
        # https://stackoverflow.com/questions/24706125/setting-a-fixed-size-for-points-in-legend
        if len(legend.legendHandles) >= 3:
            legend.legendHandles[0]._sizes = [30]
            legend.legendHandles[1]._sizes = [30]
            legend.legendHandles[2]._sizes = [30]

        plt.savefig(str(paths_for_test.build_dir / pathlib.Path('ypr.png')),
                    dpi=128,
                    pad_inches=.5,
                    bbox_inches='tight')

    except ImportError:
        print('Install matplotlib to get a pretty graph in the test output.')



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
#                                       (@@@@%%%%%%%%%&@@&.
#                              /%&&%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%&@@(
#                              *@&%%%%%%%%%&&%%%%%%%%%%%%%%%%%%&&&%%%%%%%
#                               @   @@@(@@@@%%%%%%%%%%%%%%%%&@@&* @@@   .
#                               ,   .        .  .@@@&                   /
#                                .       .                              *
#                               @@              .                       @
#                              @&&&&&&@. .    .                     *@%&@
#                              &&&&&&&&&&&&&&&&@@        *@@############@
#                     *&/ @@ #&&&&&&&&&&&&&&&&&&&&@  ###################*
#                              @&&&&&&&&&&&&&&&&&&##################@
#                                 %@&&&&&&&&&&&&&&################@
#                                        @&&&&&&&&&&%#######&@%
#  nanaimo                                   (@&&&&####@@*
#
import asyncio
import re
import textwrap
import typing

import pytest

import nanaimo
import nanaimo.connections
import nanaimo.connections.uart
import nanaimo.fixtures
import nanaimo.pytest.plugin


class Fixture(nanaimo.fixtures.Fixture):
    """
    Gathers a log over a serial connection until a given pattern is matched.
    """

    fixture_name = 'nanaimo_serial_watch'
    argument_prefix = 'lw'

    def __init__(self,
                 manager: nanaimo.fixtures.FixtureManager,
                 args: nanaimo.Namespace,
                 **kwargs: typing.Any) -> None:
        super().__init__(manager, args, **kwargs)
        if 'uart_factory' in kwargs:
            self._uart_factory = typing.cast(typing.Callable[[str, int], nanaimo.connections.uart.ConcurrentUart],
                                             kwargs['uart_factory'])
        else:
            self._uart_factory = nanaimo.connections.uart.ConcurrentUart.new_default

    @classmethod
    def on_visit_test_arguments(cls, arguments: nanaimo.Arguments) -> None:
        nanaimo.connections.uart.ConcurrentUart.on_visit_test_arguments(arguments)
        arguments.add_argument('--pattern',
                               default='.*',
                               enable_default_from_environ=True,
                               help=textwrap.dedent('''
                A Python regular expression that will be matched against each line of
                serial input received. This fixture gathers input until either this
                pattern is matched or the gather times out.''').lstrip())
        arguments.add_argument('--disruption', default='\r\n', help=textwrap.dedent('''
                Characters to send periodically on the serial link in an attempt to wake
                up connected hosts and get a default response.''').lstrip())
        arguments.add_argument('--disturb-rate', type=float, help=textwrap.dedent('''
                The rate at which the watcher will input the disruption characters to try
                to get the device on the other end of the serial pipe to respond
                (in fractional seconds)''').lstrip())
        arguments.add_argument('--update-period',
                               type=float,
                               help=textwrap.dedent('''
                The rate at which a message is logged when waiting for a log time. This
                provides feedback that the fixture is still active but that the pattern
                has not yet been matched. Omit to squelch the update message.
                (This argument is in fractional seconds).
                            ''').lstrip())

    async def on_gather(self, args: nanaimo.Namespace) -> nanaimo.Artifacts:
        """
        Watch the logs until the pattern matches.

        +--------------+---------------------------+-----------------------------------------------+
        | **Returned Artifacts**                                                                   |
        +--------------+---------------------------+-----------------------------------------------+
        | key          | type                      | Notes                                         |
        +==============+===========================+===============================================+
        | match        | re.MatchObject            | The match if result_code is 0                 |
        +--------------+---------------------------+-----------------------------------------------+
        | matched_line | str                       | The full line matched if result_code is 0     |
        +--------------+---------------------------+-----------------------------------------------+
        """
        lw_port = self.get_arg_covariant_or_fail(args, 'port')
        lw_port_speed = self.get_arg_covariant_or_fail(args, 'port_speed')

        with self._uart_factory(lw_port, lw_port_speed) as monitor:

            match_future, _ = await self.gate_tasks(self._matcher(args, monitor),
                                                    None,
                                                    self._agitator(args, monitor),
                                                    self._updater(args))

        artifacts = typing.cast(nanaimo.Artifacts, match_future.result())
        self._logger.info('Found match : %s', artifacts.matched_line)
        return artifacts

    # +-----------------------------------------------------------------------+
    # | PRIVATE
    # +-----------------------------------------------------------------------+
    async def _updater(self, args: nanaimo.Namespace) -> nanaimo.Artifacts:
        lw_update_period = self.get_arg_covariant(args, 'update_period')
        if lw_update_period is not None:
            while True:
                await asyncio.sleep(lw_update_period)
                self._logger.info('still waiting...')

        return nanaimo.Artifacts()

    async def _matcher(self, args: nanaimo.Namespace,
                       monitor: nanaimo.connections.uart.ConcurrentUart) -> nanaimo.Artifacts:
        artifacts = nanaimo.Artifacts()
        lw_pattern = self.get_arg_covariant_or_fail(args, 'pattern')
        pattern = re.compile(lw_pattern)

        self._logger.info('Starting to look for pattern : %s', str(lw_pattern))
        while True:
            result = await monitor.get_line()
            result_stripped = result.rstrip()
            if len(result_stripped) > 0:
                self._logger.debug(result_stripped)
            match = pattern.search(result)
            if match:
                setattr(artifacts, 'match', match)
                setattr(artifacts, 'matched_line', result)
                break

        return artifacts

    async def _agitator(self, args: nanaimo.Namespace,
                        monitor: nanaimo.connections.uart.ConcurrentUart) -> nanaimo.Artifacts:
        lw_disturb_rate = self.get_arg_covariant(args, 'disturb_rate')
        lw_disruption = self.get_arg_covariant_or_fail(args, 'disruption')
        if lw_disturb_rate is not None:
            while True:
                await asyncio.sleep(lw_disturb_rate)
                self._logger.debug('About to disturb the uart...')
                await monitor.put_line(lw_disruption)
        return nanaimo.Artifacts()


@nanaimo.fixtures.PluggyFixtureManager.type_factory
def get_fixture_type() -> typing.Type['Fixture']:
    return Fixture


@pytest.fixture
def nanaimo_serial_watch(request: typing.Any) -> nanaimo.fixtures.Fixture:
    """
    This Nanaimo fixture allows a test to monitor a serial port until a provided pattern matches.
    The `pytest fixture <https://docs.pytest.org/en/latest/fixture.html>`_ provides
    a :class:`nanaimo.builtin.nanaimo_serial_watch.Fixture` fixture to a pytest.

    :param pytest_request: The request object passed into the pytest fixture factory.
    :type pytest_request: _pytest.fixtures.FixtureRequest
    :rtype: nanaimo.builtin.nanaimo_serial_watch.Fixture
    """
    return nanaimo.pytest.plugin.create_pytest_fixture(request, Fixture.get_canonical_name())



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
#                                       (@@@@%%%%%%%%%&@@&.
#                              /%&&%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%&@@(
#                              *@&%%%%%%%%%&&%%%%%%%%%%%%%%%%%%&&&%%%%%%%
#                               @   @@@(@@@@%%%%%%%%%%%%%%%%&@@&* @@@   .
#                               ,   .        .  .@@@&                   /
#                                .       .                              *
#                               @@              .                       @
#                              @&&&&&&@. .    .                     *@%&@
#                              &&&&&&&&&&&&&&&&@@        *@@############@
#                     *&/ @@ #&&&&&&&&&&&&&&&&&&&&@  ###################*
#                              @&&&&&&&&&&&&&&&&&&##################@
#                                 %@&&&&&&&&&&&&&&################@
#                                        @&&&&&&&&&&%#######&@%
#  nanaimo                                   (@&&&&####@@*
#
import asyncio
import concurrent.futures
import logging
import queue
import re
import threading
import types
import typing

import serial

from . import AbstractAsyncSerial, TimestampedLine


class ConcurrentUart(AbstractAsyncSerial):
    """
    Buffers serial input on another thread and provides line-oriented access
    to/from the tty via synchronized queues.
    """
    DefaultSerialTimeoutSeconds = 1
    WriteBufferEndOfTransmission = '\4'

    @classmethod
    def new_default(cls,
                    port: str,
                    baudrate: int,
                    loop: typing.Optional[asyncio.AbstractEventLoop] = None) -> 'ConcurrentUart':
        return cls(serial.Serial(port=port, baudrate=baudrate, timeout=cls.DefaultSerialTimeoutSeconds), loop)

    def __init__(self,
                 serial_port: serial.Serial,
                 loop: typing.Optional[asyncio.AbstractEventLoop] = None,
                 eol: str = '\r\n',
                 echo: bool = False) -> None:
        super().__init__(loop)
        self._s = serial_port
        self._echo = echo
        self._eol = eol
        self._executor = concurrent.futures.ThreadPoolExecutor(max_workers=2)
        self._serial_futures = []  # type: typing.List[concurrent.futures.Future]
        self._logger_tx = logging.getLogger(type(self).__name__ + "_tx")
        self._logger_rx = logging.getLogger(type(self).__name__ + "_rx")
        self._extra_verbose = False

    @property
    def serial_port(self) -> serial.Serial:
        return self._s

    @property
    def eol(self) -> str:
        return self._eol

    @eol.setter
    def eol(self, eol: str) -> None:
        self._eol = eol

    @property
    def echo(self) -> bool:
        return self._echo

    @echo.setter
    def echo(self, value: bool) -> None:
        self._echo = value

    @property
    def timeout_seconds(self) -> float:
        return float(self._s.timeout)

    @timeout_seconds.setter
    def timeout_seconds(self, value: float) -> None:
        self._s.timeout = value

    @property
    def extra_verbose(self) -> bool:
        return self._extra_verbose

    @extra_verbose.setter
    def extra_verbose(self, extra_verbose: bool) -> None:
        self._extra_verbose = extra_verbose

    # +-----------------------------------------------------------------------+
    # | CONTEXT MANAGER
    # +-----------------------------------------------------------------------+

    def __enter__(self) -> 'ConcurrentUart':
        if not self._s.is_open:
            self._s.open()
        self._serial_futures.append(self._executor.submit(self._buffer_input))
        self._serial_futures.append(self._executor.submit(self._buffer_output))

        return self

    def __exit__(self,
                 exception_type: typing.Optional[typing.Any],
                 exception_value: typing.Optional[typing.Any],
                 traceback: typing.Optional[types.TracebackType]) -> None:
        self.stop()
        self._s.flush()
        self._s.cancel_read()
        self._write_buffer.put_nowait(self.WriteBufferEndOfTransmission)
        for serial_future in reversed(self._serial_futures):
            if not serial_future.done():
                serial_future.result((self._s.timeout if self._s.timeout > 0 else None))
        self._s.close()
        self._executor.shutdown(wait=True)

    # +-----------------------------------------------------------------------+
    # | CONCURRENT OPERATIONS
    # +-----------------------------------------------------------------------+

    def readline(self) -> typing.Optional[TimestampedLine]:
        """
        Get a line of text from the receive buffers.

        :returns: A line of text with the time it was received at.
        :rtype: TimestampedLine
        """
        try:
            return self._read_buffer.get_nowait()
        except queue.Empty:
            return None

    def writeline(self, input_line: str, end: typing.Optional[str] = None) -> float:
        """
        Enqueue a line of text to be transmitted on the serial device.

        :param str input_line: The line to put.
        :param end: Line ending (overrides the default :meth:`eol`).
        :type end: typing.Optional[str]
        :return: The monotonic system time that the line was put into the serial buffers at (see :meth:`time`).
        """
        try:
            self._write_buffer.put_nowait(input_line + (end if end is not None else self._eol))
            return self.time()
        except queue.Full:
            return False

    # +-----------------------------------------------------------------------+
    # | PRIVATE
    # +-----------------------------------------------------------------------+

    def _buffer_input(self) -> None:
        try:
            local_storage = threading.local()
            local_storage.line_buffer = ''
            while self._queues_are_running:
                self._buffer_input_step(local_storage)
        finally:
            if self._queues_are_running:
                self._logger_rx.error("read thread exiting.")
            self._queues_are_running = False

    def _buffer_input_step(self, local_storage: threading.local) -> None:
        raw_input = self._s.read()
        rx_timestamp_seconds = self.time()
        decoded_input = local_storage.line_buffer + self._rx_decoder.decode(raw_input)
        local_storage.line_buffer = ''
        decoded_lines = decoded_input.split(self._eol)
        if not decoded_input.endswith(self._eol) and len(decoded_lines) > 0:
            # last bit didn't yet have a terminator. Buffer it for the next
            # go around
            local_storage.line_buffer = decoded_lines[-1]
            decoded_lines = decoded_lines[:-1]
        for decoded_line in decoded_lines:
            try:
                timestamped_line = TimestampedLine.create(decoded_line, rx_timestamp_seconds)
                self._read_buffer.put_nowait(timestamped_line)
                if self._extra_verbose:
                    self._logger_rx.debug(re.sub('\\r', '<cr>', timestamped_line))
            except queue.Full:
                self._logger_rx.warning("read buffer overflow.")
                self._rx_buffer_overflows += 1

    def _buffer_output(self) -> None:
        try:
            while self._queues_are_running:
                self._buffer_output_step()
        finally:
            if self._queues_are_running:
                self._logger_tx.error("write thread exiting.")
            self._queues_are_running = False

    def _buffer_output_step(self) -> None:
        try:
            writeline = self._write_buffer.get(block=True)
            if writeline != self.WriteBufferEndOfTransmission:
                self._s.write(self._tx_encoder.encode(writeline))
                if self._echo:
                    self._logger_tx.info(re.sub('\\r', '<cr>', writeline))
        except queue.Empty:
            pass



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
import argparse
import asyncio
import pathlib
import typing

import pytest

import nanaimo
import nanaimo.builtin
import nanaimo.fixtures
from nanaimo import assert_success
from nanaimo.builtin import nanaimo_bar, nanaimo_cmd, nanaimo_gather


class CanonicallyNamed(nanaimo.fixtures.Fixture):

    fixture_name = 'foo_bar'


def test_canonical_name(dummy_nanaimo_fixture: nanaimo.fixtures.Fixture) -> None:
    """
    Verify the behavior of Fixture.get_canonical_name()
    """
    assert 'material.DummyFixture' == type(dummy_nanaimo_fixture).get_canonical_name()
    assert CanonicallyNamed.fixture_name == CanonicallyNamed.get_canonical_name()


@pytest.mark.asyncio
async def test_gather_coroutines(nanaimo_fixture_manager: nanaimo.fixtures.FixtureManager) -> None:

    parser = argparse.ArgumentParser()
    nanaimo_gather.Fixture.on_visit_test_arguments(nanaimo.Arguments(parser, required_prefix='gather'))
    args = nanaimo.Namespace(parser.parse_args(['--gather-coroutine', 'nanaimo_bar',
                                                '--gather-coroutine', 'nanaimo_bar']))

    gather_fixture = nanaimo_gather.Fixture(nanaimo_fixture_manager, args)

    results = await gather_fixture.gather()

    assert results.result_code == 0


@pytest.mark.timeout(10)
@pytest.mark.asyncio
async def test_observe_tasks(dummy_nanaimo_fixture: nanaimo.fixtures.Fixture) -> None:
    """
    Test the observe_tasks method of Fixture
    """

    async def evaluating() -> int:
        return 0

    async def running() -> int:
        waits = 2
        while waits > 0:
            await asyncio.sleep(.1)
            waits -= 1
        return 1

    result = await dummy_nanaimo_fixture.observe_tasks_assert_not_done(evaluating(),
                                                                       None,
                                                                       running())
    assert len(result) == 1
    should_be_running = result.pop()

    assert not should_be_running.done()

    assert 1 == await should_be_running


@pytest.mark.timeout(10)
@pytest.mark.asyncio
async def test_observe_tasks_failure(dummy_nanaimo_fixture: nanaimo.fixtures.Fixture) -> None:
    """
    Test the observe_tasks method of Fixture where the running tasks exit.
    """

    async def evaluating() -> int:
        waits = 2
        while waits > 0:
            await asyncio.sleep(.1)
            waits -= 1
        return 1

    async def running() -> int:
        return 1

    with pytest.raises(nanaimo.AssertionError):
        await dummy_nanaimo_fixture.observe_tasks_assert_not_done(evaluating(),
                                                                  None,
                                                                  running())


@pytest.mark.timeout(10)
@pytest.mark.asyncio
async def test_observe_tasks_failure_no_assert(dummy_nanaimo_fixture: nanaimo.fixtures.Fixture) -> None:
    """
    Test the observe_tasks method of Fixture where the running tasks exit but without throwing
    an assertion error.
    """

    async def evaluating() -> int:
        waits = 2
        while waits > 0:
            await asyncio.sleep(.1)
            waits -= 1
        return 1

    async def running() -> int:
        return 1

    result = await dummy_nanaimo_fixture.observe_tasks(evaluating(),
                                                       None,
                                                       running())

    assert 0 == len(result)


@pytest.mark.timeout(10)
@pytest.mark.asyncio
async def test_observe_tasks_timeout(dummy_nanaimo_fixture: nanaimo.fixtures.Fixture) -> None:
    """
    Test the observe_tasks method of Fixture where the running tasks do not exit.
    """

    async def evaluating() -> int:
        while True:
            await asyncio.sleep(1)

    async def running() -> int:
        while True:
            await asyncio.sleep(1)

    with pytest.raises(asyncio.TimeoutError):
        await dummy_nanaimo_fixture.observe_tasks_assert_not_done(evaluating(),
                                                                  1,
                                                                  running())


@pytest.mark.timeout(20)
@pytest.mark.asyncio
async def test_countdown_sleep(dummy_nanaimo_fixture: nanaimo.fixtures.Fixture) -> None:
    """
    Test the observe_tasks method of Fixture where the running tasks do not exit.
    """
    await dummy_nanaimo_fixture.countdown_sleep(5.3)


@pytest.mark.timeout(20)
@pytest.mark.asyncio
async def test_gather_timeout(gather_timeout_fixture: nanaimo.fixtures.Fixture) -> None:
    """
    Test the standard fixture timeout.
    """
    gather_timeout_fixture.gather_timeout_seconds = 1.0
    with pytest.raises(asyncio.TimeoutError):
        await gather_timeout_fixture.gather()


@pytest.mark.asyncio
async def test_subprocess_fixture_logfile(build_output: pathlib.Path) -> None:
    """
    Verify the subprocess fixture base class properly creates a logfile.
    """
    import nanaimo.version

    logfile = build_output / pathlib.Path('test_subprocess_fixture_logfile').with_suffix('.log')
    class SubprocessTestHarness(nanaimo.fixtures.SubprocessFixture):

        argument_prefix = 'test-subprocess-fixture'

        @classmethod
        def on_visit_test_arguments(cls, arguments: nanaimo.Arguments) -> None:
            super().on_visit_test_arguments(arguments)

        def on_construct_command(self, arguments: nanaimo.Namespace, inout_artifacts: nanaimo.Artifacts) -> str:
            return 'nait --version'

    subject = SubprocessTestHarness(nanaimo.fixtures.FixtureManager())
    artifacts = await subject.gather(
        test_subprocess_fixture_logfile=str(logfile),
        test_subprocess_fixture_logfile_amend=False
    )

    assert artifacts.result_code == 0

    found = False
    with open(str(artifacts.logfile), 'r') as logfile_io:
        for line in logfile_io:
            print(line)
            if line.find(nanaimo.version.__version__) != -1:
                found = True
                break

    assert found


@pytest.mark.asyncio
async def test_subprocess_fixture_environment(build_output: pathlib.Path) -> None:
    """
    Verify that the subprocess environment is populated from configuration.
    """
    class SubprocessTestHarness(nanaimo.fixtures.SubprocessFixture):

        @classmethod
        def on_visit_test_arguments(cls, arguments: nanaimo.Arguments) -> None:
            super().on_visit_test_arguments(arguments)

        def on_construct_command(self, arguments: nanaimo.Namespace, inout_artifacts: nanaimo.Artifacts) -> str:
            # NANAIMO_UNITTEST is set in our setup.cfg
            return 'echo ${NANAIMO_UNITTEST}'

    subject = SubprocessTestHarness(nanaimo.fixtures.FixtureManager())
    filter = nanaimo.fixtures.SubprocessFixture.SubprocessMessageAccumulator()
    subject.stdout_filter = filter

    artifacts = await subject.gather()

    assert artifacts.result_code == 0

    defaults = nanaimo.config.ArgumentDefaults.create_defaults_with_early_rc_config()

    nanaimo_environ = nanaimo.config.ArgumentDefaults.as_dict(defaults['environ'])
    assert 'NANAIMO_UNITTEST' in nanaimo_environ
    assert nanaimo_environ['NANAIMO_UNITTEST'] == filter.getvalue()


@pytest.mark.asyncio
@pytest.mark.skip(reason='https://github.com/thirtytwobits/nanaimo/issues/84')
async def test_composite_fixture(event_loop: asyncio.AbstractEventLoop) -> None:
    """
    Test creation of a composite fixture.
    """

    class Composite(nanaimo_bar.Fixture, nanaimo_cmd.Fixture):

        fixture_name = 'test_composite_fixture'
        argument_prefix = 'tcf'

        def __init__(self, manager: nanaimo.fixtures.FixtureManager,
                     args: nanaimo.Namespace,
                     **kwargs: typing.Any) -> None:
            nanaimo_bar.Fixture.__init__(self, manager, args, **kwargs)
            nanaimo_cmd.Fixture.__init__(self, manager, args, **kwargs)
            self._gather = nanaimo_gather.Fixture(manager, args, **kwargs)

        @classmethod
        def on_visit_test_arguments(cls, arguments: nanaimo.Arguments) -> None:
            nanaimo_bar.Fixture.visit_test_arguments(arguments)
            nanaimo_cmd.Fixture.visit_test_arguments(arguments)

        async def on_gather(self, args: nanaimo.Namespace) -> nanaimo.Artifacts:
            coroutines = [
                nanaimo_bar.Fixture.on_gather(self, args),
                nanaimo_cmd.Fixture.on_gather(self, args)
            ]
            return await self._gather.gather(gather_coroutine=coroutines)

    composite = Composite(nanaimo.fixtures.FixtureManager(),
                          nanaimo.Namespace(),
                          loop=event_loop)

    filter = nanaimo.fixtures.SubprocessFixture.SubprocessMessageAccumulator()
    composite.stdout_filter = filter

    results = await composite.gather(tcf_shell='nait --version')

    results.eat()
    assert filter.getvalue() == nanaimo.version.__version__


@pytest.mark.asyncio
async def test_subprocess_read_illegal_encoding(paths_for_test: typing.Any) -> None:
    """
    Verify the subprocess fixture base class can handle unknown input without 'sploding.
    """
    import nanaimo.version

    class SubprocessTestHarness(nanaimo.fixtures.SubprocessFixture):

        argument_prefix = 'test-subprocess-fixture'

        @classmethod
        def on_visit_test_arguments(cls, arguments: nanaimo.Arguments) -> None:
            super().on_visit_test_arguments(arguments)

        def on_construct_command(self, arguments: nanaimo.Namespace, inout_artifacts: nanaimo.Artifacts) -> str:
            return 'export PYTHONIOENCODING=utf_16_be && python {} {}'.format(
                str(paths_for_test.unicode_py), str(paths_for_test.unicode_big))

    stdout = nanaimo.fixtures.SubprocessFixture.SubprocessMessageAccumulator()
    stderr = nanaimo.fixtures.SubprocessFixture.SubprocessMessageAccumulator()
    subject = SubprocessTestHarness(nanaimo.fixtures.FixtureManager())
    subject.stdout_filter = stdout
    subject.stderr_filter = stderr
    assert_success(await subject.gather())



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#

import nanaimo
import nanaimo.fixtures


def test_ykush_exists(nanaimo_instr_ykush: nanaimo.fixtures.Fixture) -> None:
    """
    Just making sure the fixture exists.
    """
    assert nanaimo_instr_ykush.get_canonical_name() is not None



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#

import configparser
import pathlib

import nanaimo
from nanaimo.config import ArgumentDefaults


def test_create_orphan() -> None:
    """
    Create an orphan Namespace
    """
    subject = nanaimo.Namespace()
    assert 'foo' not in subject


def test_create_single_parent() -> None:
    """
    Create a namespace with a single parent.
    """
    parent = nanaimo.Namespace()
    setattr(parent, 'foo', 1)
    child = nanaimo.Namespace(parent)
    assert 1 == child.foo
    assert 'bar' not in child


def test_create_grandparent() -> None:
    """
    Create a namespace with a grand parent.
    """
    grandparent = nanaimo.Namespace()
    setattr(grandparent, 'foo', 1)
    setattr(grandparent, 'bar', 8)
    parent = nanaimo.Namespace(grandparent)
    setattr(parent, 'foo', 2)
    child = nanaimo.Namespace(parent)
    assert 2 == child.foo
    assert 8 == child.bar
    assert 1 == grandparent.foo
    assert 8 == child.bar


def test_merge() -> None:
    """
    Verify the merge method of Namespace.
    """
    oldparent = nanaimo.Namespace()
    setattr(oldparent, 'oldparent', 'yes')
    setattr(oldparent, 'name', 'oldparent')

    child = nanaimo.Namespace(oldparent)

    setattr(child, 'name', 'child')

    subject = child.merge(name='subject', merged='yes')  # type: nanaimo.Namespace

    assert 'yes' == subject.oldparent
    assert 'yes' == subject.merged
    assert 'subject' == subject.name


def test_overrides(test_config: pathlib.Path) -> None:
    """
    Verify that we can provide overrides for namespace attributes.
    """
    fake_args = nanaimo.Namespace()
    setattr(fake_args, 'rcfile', str(test_config))
    defaults = ArgumentDefaults()
    subject = nanaimo.Namespace(None, defaults)
    assert 'test_attr_yup' not in subject

    defaults.set_args(fake_args)
    setattr(subject, 'test_attr_nope', 'yup')
    assert subject.test_attr_yup == 'yup'
    assert subject.test_attr_nope == 'yup'


def test_setup_cfg(local_setup_cfg: pathlib.Path) -> None:
    """
    Verify we can embed Nanaimo configuration in a setup.cfg file.
    """

    # Quick self-check since we don't trust ourselves.
    parser = configparser.ConfigParser()
    assert parser.read([str(local_setup_cfg)]) == [str(local_setup_cfg)]
    assert parser['nanaimo']['test_cfg'] == '1'

    fake_args = nanaimo.Namespace()
    setattr(fake_args, 'rcfile', str(local_setup_cfg))
    defaults = ArgumentDefaults(fake_args)
    subject = nanaimo.Namespace(None, defaults)
    assert subject.test_cfg == '2'


def test_setup_cfg_with_merge(local_setup_cfg: pathlib.Path) -> None:
    """
    Make sure merge acts as expected.
    """
    fake_args = nanaimo.Namespace()
    setattr(fake_args, 'rcfile', str(local_setup_cfg))
    defaults = ArgumentDefaults(fake_args)
    subject = nanaimo.Namespace(None, defaults).merge()  # type: nanaimo.Namespace
    assert subject.test_cfg == '2'


def test_setup_cfg_with_none_value(local_setup_cfg: pathlib.Path) -> None:
    """
    Make sure None values are handled properly when looking for overrides.
    """
    fake_args = nanaimo.Namespace()
    setattr(fake_args, 'rcfile', str(local_setup_cfg))
    setattr(fake_args, 'test_cfg', None)
    defaults = ArgumentDefaults(fake_args)
    subject0 = nanaimo.Namespace(fake_args, defaults, allow_none_values=True)
    assert subject0.test_cfg is None
    subject1 = nanaimo.Namespace(fake_args, defaults, allow_none_values=False)
    assert subject1.test_cfg == '2'



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
#                                       (@@@@%%%%%%%%%&@@&.
#                              /%&&%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%&@@(
#                              *@&%%%%%%%%%&&%%%%%%%%%%%%%%%%%%&&&%%%%%%%
#                               @   @@@(@@@@%%%%%%%%%%%%%%%%&@@&* @@@   .
#                               ,   .        .  .@@@&                   /
#                                .       .                              *
#                               @@              .                       @
#                              @&&&&&&@. .    .                     *@%&@
#                              &&&&&&&&&&&&&&&&@@        *@@############@
#                     *&/ @@ #&&&&&&&&&&&&&&&&&&&&@  ###################*
#                              @&&&&&&&&&&&&&&&&&&##################@
#                                 %@&&&&&&&&&&&&&&################@
#                                        @&&&&&&&&&&%#######&@%
#  nanaimo                                   (@&&&&####@@*
#
"""
TODO: See https://www.learn.yepkit.com/reference/ykushcmd-reference-ykush/1/2 for the command
strings.
"""

import typing

import pytest

import nanaimo
import nanaimo.fixtures
import nanaimo.pytest.plugin


class Fixture(nanaimo.fixtures.SubprocessFixture):
    """
    Fixture for controlling Yepkit USB hubs with switchable power. For example
    the `YKUSH3 <https://www.yepkit.com/product/300110/YKUSH3>`_ is a 3-port
    USB-3 hub that allows individual control of the power rails for each port.

    This is a subprocess fixture that requires the ``ykushcmd`` program is available in the subprocess
    environment (see Yepkit's documentation for how to build this from source). All arguments can be
    overridden via the fixtures gather method. The supported commands are:

    +-----------------+--------------------------------------------------------+----------------------------+
    | command         | example                                                | Description                |
    +=================+========================================================+============================+
    | ``yku-all-on``  | ``await nanaimo_instr_ykush.gather(yku_all_on=True)``  | Turn on power to all ports |
    |                 |                                                        | on the YKUSH.              |
    +-----------------+--------------------------------------------------------+----------------------------+
    | ``yku-all-off`` | ``await nanaimo_instr_ykush.gather(yku_all_off=True)`` | Turn off power to all      |
    |                 |                                                        | ports on the YKUSH.        |
    +-----------------+--------------------------------------------------------+----------------------------+
    | ``yku-command`` | ``await nanaimo_instr_ykush.gather(yku_command='-l')`` | Pass-through any command   |
    |                 |                                                        | to ykushcmd.               |
    +-----------------+--------------------------------------------------------+----------------------------+
    """

    fixture_name = 'nanaimo_ykush'
    argument_prefix = 'yku'

    ykush_cmd = 'ykushcmd'

    @classmethod
    def on_visit_test_arguments(cls, arguments: nanaimo.Arguments) -> None:
        arguments.add_argument('--model',
                               default='ykush3',
                               help='The ykush board type.')
        arguments.add_argument('--serial',
                               help='A serial number of the board to send the command to.')
        arguments.add_argument('--command',
                               help='Simple pass through of arguments to {}'.format(cls.ykush_cmd))
        arguments.add_argument('--all-on',
                               action='store_true',
                               help='Turn on power to all ports (--{}-command will be ignored)'
                               .format(cls.argument_prefix))
        arguments.add_argument('--all-off',
                               action='store_true',
                               help='Turn on power to all ports (--{}-command will be ignored)'
                               .format(cls.argument_prefix))

    def on_construct_command(self, arguments: nanaimo.Namespace, inout_artifacts: nanaimo.Artifacts) -> str:
        """
        .. invisible-code-block: python

            from nanaimo.fixtures import FixtureManager
            from nanaimo import Namespace, Arguments, Artifacts
            from nanaimo.instruments import ykush

            import os
            import asyncio

            loop = asyncio.get_event_loop()
            manager = FixtureManager(loop=loop)

            cmd = ykush.Fixture(manager)

            ns = Namespace()
            artifacts = Artifacts()

            command = cmd.on_construct_command(ns, artifacts)
            assert command.find('-h') > 0

            setattr(ns, '{}_all_off'.format(ykush.Fixture.get_argument_prefix()), True)
            command = cmd.on_construct_command(ns, artifacts)
            assert command.find('-d a') > 0
            setattr(ns, '{}_all_off'.format(ykush.Fixture.get_argument_prefix()), None)

            setattr(ns, '{}_all_on'.format(ykush.Fixture.get_argument_prefix()), True)
            command = cmd.on_construct_command(ns, artifacts)
            assert command.find('-u a') > 0

            # off takes precedence
            setattr(ns, '{}_all_off'.format(ykush.Fixture.get_argument_prefix()), True)
            command = cmd.on_construct_command(ns, artifacts)
            assert command.find('-d a') > 0
            setattr(ns, '{}_all_off'.format(ykush.Fixture.get_argument_prefix()), None)
            setattr(ns, '{}_all_on'.format(ykush.Fixture.get_argument_prefix()), None)

            setattr(ns, '{}_command'.format(ykush.Fixture.get_argument_prefix()), 'foo')
            command = cmd.on_construct_command(ns, artifacts)
            assert command.find('foo') > 0

        """
        serial_arg = self.get_arg_covariant(arguments, 'serial')
        serial = ('' if serial_arg is None else ' -s ' + serial_arg)

        all_on = self.get_arg_covariant(arguments, 'all_on', False)
        all_off = self.get_arg_covariant(arguments, 'all_off', False)
        pass_thru = self.get_arg_covariant(arguments, 'command')

        if all_off:
            self._logger.debug('Turning all ports off.')
            command = '-d a'
        elif all_on:
            self._logger.debug('Turning all ports on.')
            command = '-u a'
        elif pass_thru is not None:
            command = str(pass_thru)
            self._logger.debug('Passing through command {}'.format(pass_thru))
        else:
            command = '-h'

        return '{ykush_cmd} {board}{serial} {command}'.format(
            ykush_cmd=self.ykush_cmd,
            board=self.get_arg_covariant(arguments, 'model', ''),
            serial=serial,
            command=command
        )


@nanaimo.fixtures.PluggyFixtureManager.type_factory
def get_fixture_type() -> typing.Type['Fixture']:
    return Fixture


@pytest.fixture
def nanaimo_instr_ykush(request: typing.Any) -> nanaimo.fixtures.Fixture:
    """
    Provides a :class:`nanaimo.instruments.ykush.Fixture` fixture to a pytest.
    This fixture controls a `YKUSH <https://www.yepkit.com/products/ykush>`_
    family board attached to the system via USB.

    :param pytest_request: The request object passed into the pytest fixture factory.
    :type pytest_request: _pytest.fixtures.FixtureRequest
    :return: A fixture providing control of a YKUSH usb hub.
    :rtype: nanaimo.instruments.ykush.Fixture
    """
    return nanaimo.pytest.plugin.create_pytest_fixture(request, Fixture.get_canonical_name())



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
#                                       (@@@@%%%%%%%%%&@@&.
#                              /%&&%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%&@@(
#                              *@&%%%%%%%%%&&%%%%%%%%%%%%%%%%%%&&&%%%%%%%
#                               @   @@@(@@@@%%%%%%%%%%%%%%%%&@@&* @@@   .
#                               ,   .        .  .@@@&                   /
#                                .       .                              *
#                               @@              .                       @
#                              @&&&&&&@. .    .                     *@%&@
#                              &&&&&&&&&&&&&&&&@@        *@@############@
#                     *&/ @@ #&&&&&&&&&&&&&&&&&&&&@  ###################*
#                              @&&&&&&&&&&&&&&&&&&##################@
#                                 %@&&&&&&&&&&&&&&################@
#                                        @&&&&&&&&&&%#######&@%
#  nanaimo                                   (@&&&&####@@*
#
"""
Built-in :class:`nanaimo.fixtures.Fixture` objects for common scenarios. See individual fixture documentation for use.
"""
import asyncio
import pathlib
import typing

import pytest

import nanaimo
import nanaimo.connections
import nanaimo.connections.uart
import nanaimo.fixtures
import nanaimo.instruments
import nanaimo.instruments.jlink
import nanaimo.parsers
import nanaimo.parsers.gtest
import nanaimo.pytest.plugin


class Fixture(nanaimo.fixtures.Fixture):
    """
    Uploads a firmware using JLink and monitors a UART expected gtest output. Returns 0 if all
    gtest tests pass else returns non-zero.
    """
    fixture_name = 'gtest_over_jlink'
    argument_prefix = 'gtest'

    @classmethod
    def on_visit_test_arguments(cls, arguments: nanaimo.Arguments) -> None:
        nanaimo.connections.uart.ConcurrentUart.on_visit_test_arguments(arguments)
        nanaimo.instruments.jlink.ProgramUploaderJLink.on_visit_test_arguments(arguments)
        arguments.add_argument('--timeout',
                               default=30.0,
                               type=float,
                               help='Time to wait for the gtest to complete in seconds.')

    async def on_gather(self, args: nanaimo.Namespace) -> nanaimo.Artifacts:

        gtest_timeout = self.get_arg_covariant(args, 'timeout', None)
        gtest_port = self.get_arg_covariant_or_fail(args, 'port')
        gtest_port_speed = self.get_arg_covariant_or_fail(args, 'port_speed')
        gtest_upload_timeout_seconds = self.get_arg_covariant(args, 'upload_timeout_seconds')
        gtest_base_path = self.get_arg_covariant_or_fail(args, 'base_path')
        gtest_jlink_scripts = self.get_arg_covariant_or_fail(args, 'jlink_scripts')

        uploader = nanaimo.instruments.jlink.ProgramUploaderJLink()
        jlink_scripts = pathlib.Path(gtest_base_path).glob(gtest_jlink_scripts)
        parser = nanaimo.parsers.gtest.Parser(gtest_timeout)

        result = 0
        with nanaimo.connections.uart.ConcurrentUart.new_default(gtest_port, gtest_port_speed) as monitor:
            for script in jlink_scripts:
                if result != 0:
                    break
                result = await asyncio.wait_for(uploader.upload(script), timeout=gtest_upload_timeout_seconds)
                if result != 0:
                    break
                result = await parser.read_test(monitor)

        return nanaimo.Artifacts(result)


@nanaimo.fixtures.PluggyFixtureManager.type_factory
def get_fixture_type() -> typing.Type['Fixture']:
    return Fixture


@pytest.fixture
def gtest_over_jlink(request: typing.Any) -> nanaimo.fixtures.Fixture:
    return nanaimo.pytest.plugin.create_pytest_fixture(request, Fixture.get_canonical_name())



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#

import argparse
import asyncio
import os
import pathlib
import typing
from unittest.mock import MagicMock

import pytest

import material
import nanaimo
import nanaimo.config
import nanaimo.connections
import nanaimo.connections.uart
import nanaimo.instruments.jlink
import nanaimo.parsers.gtest
from nanaimo import set_subprocess_environment


@pytest.mark.timeout(10)
def test_uart_monitor(serial_simulator_type: typing.Type) -> None:
    """
    Verify the nanaimo.ConcurrentUart class using a mock serial port.
    """
    serial = serial_simulator_type(material.FAKE_TEST_SUCCESS)
    last_line = material.FAKE_TEST_SUCCESS[-1]
    with nanaimo.connections.uart.ConcurrentUart(serial) as monitor:
        while True:
            line = monitor.readline()
            if line is None:
                os.sched_yield()
                continue
            elif line == last_line:
                break


@pytest.mark.asyncio
async def test_program_uploader(mock_JLinkExe: pathlib.Path,
                                s32K144_jlink_script: pathlib.Path) -> None:
    uploader = nanaimo.instruments.jlink.ProgramUploaderJLink(mock_JLinkExe)
    assert 0 == await uploader.upload(s32K144_jlink_script)


@pytest.mark.asyncio
async def test_program_uploader_failure(mock_JLinkExe: pathlib.Path,
                                        s32K144_jlink_script: pathlib.Path) -> None:
    uploader = nanaimo.instruments.jlink.ProgramUploaderJLink(mock_JLinkExe, ['--simulate-error'])
    assert 0 != await uploader.upload(s32K144_jlink_script)


@pytest.mark.asyncio
@pytest.mark.timeout(60)
async def test_program_while_monitoring(mock_JLinkExe: pathlib.Path,
                                        s32K144_jlink_scripts: typing.Iterator[pathlib.Path],
                                        serial_simulator_type: typing.Type) -> None:
    scripts = s32K144_jlink_scripts
    uploader = nanaimo.instruments.jlink.ProgramUploaderJLink(mock_JLinkExe)
    serial = serial_simulator_type(material.FAKE_TEST_SUCCESS)
    uploads = 0
    with nanaimo.connections.uart.ConcurrentUart(serial) as monitor:
        for script in scripts:
            serial.reset_fake_input()
            results = await asyncio.gather(
                nanaimo.parsers.gtest.Parser(10).read_test(monitor),
                uploader.upload(script)
            )
            assert 2 == len(results)

            for result in results:
                assert 0 == result
            uploads += 1
    assert uploads > 1


@pytest.mark.asyncio
async def test_failed_test(serial_simulator_type: typing.Type) -> None:
    serial = serial_simulator_type(material.FAKE_TEST_FAILURE)
    with nanaimo.connections.uart.ConcurrentUart(serial) as monitor:
        assert 1 == await nanaimo.parsers.gtest.Parser(10).read_test(monitor)


@pytest.mark.asyncio
async def test_timeout_while_monitoring(serial_simulator_type: typing.Type) -> None:
    serial = serial_simulator_type(['gibberish'], loop_fake_data=False)
    with nanaimo.connections.uart.ConcurrentUart(serial) as monitor:
        assert 0 != await nanaimo.parsers.gtest.Parser(4.0).read_test(monitor)


def test_enable_default_from_environ(nanaimo_defaults: nanaimo.config.ArgumentDefaults) -> None:
    a = nanaimo.Arguments(argparse.ArgumentParser(), nanaimo_defaults)

    a.add_argument('yep', enable_default_from_environ=False)

    with pytest.raises(ValueError):
        a.add_argument('nope', enable_default_from_environ=True)

    with pytest.raises(ValueError):
        a.add_argument('-n', enable_default_from_environ=True)

    a.add_argument('--yep', enable_default_from_environ=True)

    with pytest.raises(RuntimeError):
        a.add_argument('--yep', enable_default_from_environ=True)


@pytest.mark.parametrize('required_prefix,test_positional_args,test_expected_args',
                         [  # type: ignore
                             ('tt', ['--foo-bar'], ['--tt-foo-bar']),
                             ('z', ['-f', '--foo-bar'], ['-f', '--z-foo-bar']),
                             ('pre', ['--foo-bar', '-x'], ['--pre-foo-bar', '-x']),
                             ('a', ['-foo-bar', '-x'], ['-foo-bar', '-x'])
                         ])
def test_require_prefix(required_prefix, test_positional_args, test_expected_args) -> None:
    parser = MagicMock(spec=argparse.ArgumentParser)
    parser.add_argument = MagicMock()

    a = nanaimo.Arguments(parser, required_prefix=required_prefix)

    a.add_argument(*test_positional_args)

    parser.add_argument.assert_called_once_with(*test_expected_args)


def test_set_subprocess_environment_no_environ() -> None:
    """
    Verify that no exceptions are thrown if the defaults config lacks an ``environ`` key.
    """
    defaults = MagicMock(spec=nanaimo.config.ArgumentDefaults)
    defaults.__getitem__ = MagicMock(side_effect=KeyError())

    set_subprocess_environment(nanaimo.Namespace(defaults=defaults))


def test_get_as_merged_dict() -> None:
    """
    Verify that no exceptions are thrown if the defaults config lacks an ``environ`` key
    when using Namespace.get_as_merged_dict()
    """
    defaults = MagicMock(spec=nanaimo.config.ArgumentDefaults)
    defaults.__getitem__ = MagicMock(side_effect=KeyError())

    nanaimo.Namespace(defaults=defaults).get_as_merged_dict('environ')



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#

import typing

import pytest

import nanaimo
import nanaimo.builtin
import nanaimo.builtin.nanaimo_serial_watch
import nanaimo.connections
import nanaimo.connections.uart


@pytest.mark.asyncio
async def test_default_match(nanaimo_serial_watch: nanaimo.builtin.nanaimo_serial_watch.Fixture,
                             serial_simulator_type: typing.Type) -> None:
    """
    The default match matches everything in the first line.
    """
    simulated_serial = serial_simulator_type(['one', 'two'])

    def uart_factory(*args: typing.Any) -> nanaimo.connections.uart.ConcurrentUart:
        return nanaimo.connections.uart.ConcurrentUart(simulated_serial)
    nanaimo_serial_watch._uart_factory = uart_factory
    artifacts = await nanaimo_serial_watch.gather(lw_port='foo')
    assert artifacts.matched_line == 'one'


@pytest.mark.asyncio
async def test_realistic_match(nanaimo_serial_watch: nanaimo.builtin.nanaimo_serial_watch.Fixture,
                               serial_simulator_type: typing.Type) -> None:
    """
    Now test a more realistic matching scenario.
    """
    line_to_match = 'FOO LINUX Distribution 7.5'
    simulated_serial = serial_simulator_type([
        'UAVCAN is an open lightweight protocol designed for reliable',
        'communication in aerospace and robotic applications',
        'via robust vehicle bus networks.',

        line_to_match,

        'Features:',
        '- Democratic network – no bus master, no single point of',
        'failure.',
        '- Publish/subscribe and request/response (RPC1)',
        'exchange semantics.'
    ])

    def uart_factory(*args: typing.Any) -> nanaimo.connections.uart.ConcurrentUart:
        return nanaimo.connections.uart.ConcurrentUart(simulated_serial)
    nanaimo_serial_watch._uart_factory = uart_factory
    artifacts = await nanaimo_serial_watch.gather(lw_port='foo', lw_pattern=r'LINUX\s+Distribution\s+(\d+\.\d+)')
    assert artifacts.matched_line == line_to_match
    assert artifacts.match.group(1) == '7.5'



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
#                                       (@@@@%%%%%%%%%&@@&.
#                              /%&&%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%&@@(
#                              *@&%%%%%%%%%&&%%%%%%%%%%%%%%%%%%&&&%%%%%%%
#                               @   @@@(@@@@%%%%%%%%%%%%%%%%&@@&* @@@   .
#                               ,   .        .  .@@@&                   /
#                                .       .                              *
#                               @@              .                       @
#                              @&&&&&&@. .    .                     *@%&@
#                              &&&&&&&&&&&&&&&&@@        *@@############@
#                     *&/ @@ #&&&&&&&&&&&&&&&&&&&&@  ###################*
#                              @&&&&&&&&&&&&&&&&&&##################@
#                                 %@&&&&&&&&&&&&&&################@
#                                        @&&&&&&&&&&%#######&@%
#  nanaimo                                   (@&&&&####@@*
#
"""
Nanaimo provides a collection of pytest fixtures all defined in this module.

.. invisible-code-block: python
    import nanaimo
    import nanaimo.fixtures
    import pytest
    import typing

.. code-block:: python

    # In my_namespace/__init__.py

    class MyTestFixture(nanaimo.fixtures.Fixture):

        @classmethod
        def on_visit_test_arguments(cls, arguments: nanaimo.Arguments) -> None:
            pass

        async def on_gather(self, args: nanaimo.Namespace) -> nanaimo.Artifacts:
            artifacts = nanaimo.Artifacts()
            # Do your on-target testing here and store results in nanaimo.Artifacts.
            return artifacts


    @nanaimo.fixtures.PluggyFixtureManager.type_factory
    def get_fixture_type() -> typing.Type['nanaimo.fixtures.Fixture']:
        return MyTestFixture

Individual fixtures can choose to present pytest fixtures directly using `pytest.fixture`
and :func:`create_pytest_fixture`

.. code-block:: python

    @pytest.fixture
    def my_test_fixture(request: typing.Any) -> nanaimo.fixtures.Fixture:
        return nanaimo.pytest.plugin.create_pytest_fixture(request, MyTestFixture.get_canonical_name())


In your setup.cfg you'll first need to register nanaimo with pytest ::

    [options]
        pytest11 =
            pytest_nanaimo = nanaimo.pytest.plugin

then, if you do want to expose your fixture directly, you'll need to add your fixture's namespace::

    [options]
        pytest11 =
            pytest_nanaimo = nanaimo.pytest.plugin
            pytest_my_plugin = my_namspace

"""
import logging
import typing

import _pytest
import pytest

import nanaimo
import nanaimo.config
import nanaimo.display
import nanaimo.fixtures

_fixture_manager = None  # type: typing.Optional[nanaimo.fixtures.FixtureManager]
"""
Pytest plugin singleton. The first time our pytest plugin is invoked
we populate to provide a peristent store of Nanaimo fixtures and artifacts.
"""


def _get_default_fixture_manager() -> nanaimo.fixtures.FixtureManager:
    global _fixture_manager
    if _fixture_manager is None:
        _fixture_manager = nanaimo.fixtures.nanaimo.fixtures.PluggyFixtureManager()

    return _fixture_manager


def create_pytest_fixture(pytest_request: typing.Any,
                          canonical_name_or_type: typing.Union[str, typing.Type['nanaimo.fixtures.Fixture']]) \
        -> nanaimo.fixtures.Fixture:
    """
    Create a fixture for a `pytest.fixture` request. This method ensures the fixture is created through
    the default :class:`FixtureManager`. For example, using :class:`nanaimo.fixtures.PluggyFixtureManager`
    you must first register your fixture like thus::

        @nanaimo.fixtures.PluggyFixtureManager.type_factory
        def get_fixture_type() -> typing.Type['nanaimo.fixtures.Fixture']:
            return MyFixtureType

    Then you can register your fixture with pytest like this::

        @pytest.fixture
        def nanaimo_pytest_my_fixture(request: typing.Any) -> nanaimo.fixtures.Fixture:
            return nanaimo.pytest.plugin.create_pytest_fixture(request, MyFixtureType.canonical_name())

    :param pytest_request: The request object passed into the pytest fixture factory.
    :type pytest_request: _pytest.fixtures.FixtureRequest
    :param canonical_name_or_type: The canonical name of the fixture to create
        (see :meth:`nanaimo.fixtures.Fixture.get_canonical_name`) or the fixture type to create.
    :return: Either a new fixture or a fixture of the same name that was already created for the default
        :class:`FixtureManager`.
    :raises KeyError: if ``fixture_type`` was not a registered nanaimo fixture and canonical_name_or_type was
        a string.
    """
    fm = _get_default_fixture_manager()
    args = pytest_request.config.option
    args_ns = nanaimo.Namespace(args, nanaimo.config.ArgumentDefaults(args), allow_none_values=False)
    canonical_name = (canonical_name_or_type if isinstance(canonical_name_or_type, str)
                      else canonical_name_or_type.get_canonical_name())
    try:
        return fm.create_fixture(canonical_name, args_ns)
    except KeyError:
        if isinstance(canonical_name_or_type, str):
            raise
    return canonical_name_or_type(fm, args_ns)


@pytest.fixture
def nanaimo_fixture_manager(request: typing.Any) -> nanaimo.fixtures.FixtureManager:
    """
    The default fixture for Nanaimo. Available as `nanaimo_fixture_manager`

    .. invisible-code-block: python

        import nanaimo
        import nanaimo.fixtures
        import pytest
        import asyncio

    .. code-block:: python

        class MyFixture(nanaimo.fixtures.Fixture):

            @classmethod
            def on_visit_test_arguments(cls, arguments: nanaimo.Arguments) -> None:
                pass

            async def on_gather(self, args: nanaimo.Namespace) -> nanaimo.Artifacts:
                return nanaimo.Artifacts()

        def test_example(nanaimo_fixture_manager: nanaimo.fixtures.FixtureManager) -> None:

            with pytest.raises(KeyError):
                # If the fixture hasn't been registered with the manager
                # this throws.
                my_fixture = nanaimo_fixture_manager.create_fixture('my_fixture')

            # Use this when creating new fixtures.
            my_new_fixture = MyFixture(nanaimo_fixture_manager)

    .. invisible-code-block: python

        class DummyFixtureManager(nanaimo.fixtures.FixtureManager):
            def create_fixture(self,
                       fixture_name,
                       args = None,
                       loop = None):
                raise KeyError()

        test_example(DummyFixtureManager())

    :param pytest_request: The request object passed into the pytest fixture factory.
    :type pytest_request: _pytest.fixtures.FixtureRequest
    :return: A reference to the default fixture manager.
    :rtype: nanaimo.fixtures.FixtureManager
    """
    return _get_default_fixture_manager()


@pytest.fixture
def nanaimo_arguments(request: typing.Any) -> nanaimo.Namespace:
    """
    Exposes the commandline arguments and defaults provided to a test.

    .. invisible-code-block: python

        import nanaimo
        import nanaimo.fixtures

    .. code-block:: python

        def test_example(nanaimo_arguments: nanaimo.Namespace) -> None:
            an_argument = nanaimo_arguments.some_arg

    :param pytest_request: The request object passed into the pytest fixture factory.
    :type pytest_request: _pytest.fixtures.FixtureRequest
    :return: A namespace with the pytest commandline args added per the documented rules.
    :rtype: nanaimo.Namespace
    """
    return nanaimo.Namespace(request.config.option)


@pytest.fixture
def nanaimo_log(request: typing.Any) -> logging.Logger:
    """
    Provides the unit tests with a logger configured for use with the Nanaimo
    framework.

    .. note ::

        For now this is just a Python logger. Future revisions may add capabilities like the
        ability to log to a display or otherwise provide feedback to humans about the status
        of a test.

    .. invisible-code-block: python

        import nanaimo
        import logging

    .. code-block:: python

        def test_example(nanaimo_log: logging.Logger) -> None:
            nanaimo_log.info('Hiya')

    It's recommended that all Nanaimo tests configure logging in a tox.ini, pytest.ini, or
    pyproject.toml (when this is supported). For example, the following section in tox.ini
    would enable cli logging for nanaimo tests::

        [pytest]
        log_cli = true
        log_cli_level = DEBUG
        log_format = %(asctime)s %(levelname)s %(name)s: %(message)s
        log_date_format = %Y-%m-%d %H:%M:%S

    :param pytest_request: The request object passed into the pytest fixture factory.
    :type pytest_request: _pytest.fixtures.FixtureRequest
    :return: A logger for use by Nanaimo tests.
    :rtype: logging.Logger
    """
    return logging.getLogger(request.function.__name__)


def assert_success(artifacts: nanaimo.Artifacts) -> nanaimo.Artifacts:
    """
    Syntactic sugar to allow more fluent handling of :meth:`fixtures.Fixture.gather`
    artifacts. For example:

    .. invisible-code-block: python

        import asyncio
        import nanaimo
        from nanaimo import Artifacts
        from nanaimo.fixtures import Fixture, FixtureManager

        _doc_loop = asyncio.new_event_loop()

        class DummyFixture(Fixture):
            @classmethod
            def on_visit_test_arguments(cls, arguments: nanaimo.Arguments) -> None:
                pass

            async def on_gather(self, args: nanaimo.Namespace) -> nanaimo.Artifacts:
                return nanaimo.Artifacts()

        fixture = DummyFixture(FixtureManager(loop=_doc_loop))

    .. code-block:: python

        from nanaimo.pytest.plugin import assert_success

        async def test_my_fixture():

            artifacts = assert_success(await fixture.gather())

            # Now we can use the artifacts. If the gather had returned
            # non-zero for the result_code an assertion error would have
            # been raised.

    .. invisible-code-block: python

        _doc_loop.run_until_complete(test_my_fixture())

    :param artifacts: The artifacts to assert on.
    :type artifacts: nanaimo.Artifacts
    :returns: artifacts (for convenience).
    :rtype: nanaimo.Artifacts()
    """
    assert artifacts.result_code == 0
    return artifacts


def assert_success_if(artifacts: nanaimo.Artifacts,
                      conditional: typing.Callable[[nanaimo.Artifacts], bool]) -> nanaimo.Artifacts:
    """
    Syntactic sugar to allow more fluent handling of :meth:`fixtures.Fixture.gather`
    artifacts but with a user-supplied conditional.

    .. invisible-code-block: python

        import asyncio
        import pytest
        import nanaimo
        from nanaimo import Artifacts
        from nanaimo.fixtures import Fixture, FixtureManager

        _doc_loop = asyncio.new_event_loop()

        class DummyFixture(Fixture):
            @classmethod
            def on_visit_test_arguments(cls, arguments: nanaimo.Arguments) -> None:
                pass

            async def on_gather(self, args: nanaimo.Namespace) -> nanaimo.Artifacts:
                a = nanaimo.Artifacts()
                setattr(a, 'foo', 'bar')
                return a

        fixture = DummyFixture(FixtureManager(loop=_doc_loop))

    .. code-block:: python

        from nanaimo.pytest.plugin import assert_success_if

        async def test_my_fixture():

            def fail_if_no_foo(artifacts: Artifacts) -> bool:
                return 'foo' in artifacts

            artifacts = assert_success_if(await fixture.gather(), fail_if_no_foo)

            print('artifacts have foo. It\'s value is {}'.format(artifacts.foo))

    .. invisible-code-block: python

        _doc_loop.run_until_complete(test_my_fixture())

        async def test_failure():

            assert_success_if(await fixture.gather(), lambda _: False)

        with pytest.raises(AssertionError):
            _doc_loop.run_until_complete(test_failure())

    :param artifacts: The artifacts to assert on.
    :type artifacts: nanaimo.Artifacts
    :param conditiona: A method called to evaluate gathered artifacts iff :data:`Artifacts.result_code` is 0.
        Return False to trigger an assertion, True to pass.
    :returns: artifacts (for convenience).
    :rtype: nanaimo.Artifacts()
    """
    assert artifacts.result_code == 0
    assert conditional(artifacts)
    return artifacts


_display_singleton = None  # type: typing.Optional[nanaimo.display.CharacterDisplay]


def _get_display() -> nanaimo.display.CharacterDisplay:
    global _display_singleton
    if _display_singleton is None:
        # TODO: #65, use generic version
        _display_singleton = typing.cast(nanaimo.display.CharacterDisplay,
                                         _get_default_fixture_manager().create_fixture('character_display'))
    return _display_singleton

# +---------------------------------------------------------------------------+
# | PYTEST HOOKS
# +---------------------------------------------------------------------------+


def pytest_addoption(parser) -> None:  # type: ignore
    """
    See :func:`_pytest.hookspec.pytest_addoption` for documentation.
    Also see the "`Writing Plugins <https://docs.pytest.org/en/latest/writing_plugins.html>`_"
    guide.
    """
    manager = _get_default_fixture_manager()
    nanaimo_defaults = nanaimo.config.ArgumentDefaults.create_defaults_with_early_rc_config()
    nanaimo_options = parser.getgroup('nanaimo')
    nanaimo_options.addoption('--environ', action='append', help='Environment variables to provide to subprocesses')
    nanaimo_arguments = nanaimo.Arguments(nanaimo_options, defaults=nanaimo_defaults, filter_duplicates=True)

    for fixture_type in manager.fixture_types():
        group = parser.getgroup(fixture_type.get_canonical_name())
        nanaimo_arguments.set_inner_arguments(group)
        fixture_type.visit_test_arguments(nanaimo_arguments)


def pytest_sessionstart(session: _pytest.main.Session) -> None:
    """
    See :func:`_pytest.hookspec.pytest_sessionstart` for documentation.
    Also see the "`Writing Plugins <https://docs.pytest.org/en/latest/writing_plugins.html>`_"
    guide.
    """
    args = session.config.option
    args_ns = nanaimo.Namespace(args, nanaimo.config.ArgumentDefaults(args), allow_none_values=False)
    nanaimo.set_subprocess_environment(args_ns)
    _get_display().set_bg_colour(0, 0, 255)


def pytest_runtest_setup(item: typing.Any) -> None:
    """
    See :func:`_pytest.hookspec.pytest_runtest_setup` for documentation.
    Also see the "`Writing Plugins <https://docs.pytest.org/en/latest/writing_plugins.html>`_"
    guide.
    """
    display = _get_display()
    display.clear(display_default_message=False)
    display.write(item.name)


def pytest_sessionfinish(session: _pytest.main.Session, exitstatus: int) -> None:
    """
    See :func:`_pytest.hookspec.pytest_sessionfinish` for documentation.
    Also see the "`Writing Plugins <https://docs.pytest.org/en/latest/writing_plugins.html>`_"
    guide.
    """
    display = _get_display()
    display.clear(display_default_message=True)
    if exitstatus == 0:
        display.set_bg_colour(0, 255, 0)
    else:
        display.set_bg_colour(255, 0, 0)



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#

import pytest

import nanaimo
import nanaimo.fixtures


def test_nanaimo_fixture_manager(nanaimo_fixture_manager: nanaimo.fixtures.FixtureManager) -> None:
    """
    Ensure the fixture manager ... er; fixture works as expected.
    """
    assert isinstance(nanaimo_fixture_manager, nanaimo.fixtures.nanaimo.fixtures.PluggyFixtureManager)
    gtest_fixture = nanaimo_fixture_manager.create_fixture('gtest_over_jlink', nanaimo.Namespace())
    assert isinstance(gtest_fixture, nanaimo.fixtures.Fixture)


@pytest.mark.asyncio
async def test_gtest_over_jlink_plugin(gtest_over_jlink: nanaimo.fixtures.Fixture) -> None:
    """
    Make sure we've properly exported gtest_over_jlink as a pytest plugin.
    """
    assert isinstance(gtest_over_jlink, nanaimo.fixtures.Fixture)


@pytest.mark.asyncio
async def test_nanaimo_bar(nanaimo_bar: nanaimo.fixtures.Fixture) -> None:
    """
    Test eating a Nanaimo bar (exercises directly exposing a plugin to pytest.)
    """
    assert nanaimo_bar is not None
    assert 'nanaimo_bar' == nanaimo_bar.name
    artifacts = await nanaimo_bar.gather()
    assert artifacts is not None
    artifacts.eat()
    assert nanaimo_bar.loop.is_running()


@pytest.mark.asyncio
async def test_another_nanaimo_bar(nanaimo_fixture_manager: nanaimo.fixtures.FixtureManager) -> None:
    """
    Test eating another Nanaimo bar (exercises using fixtures across multiple tests)
    """
    nanaimo_bar = nanaimo_fixture_manager.create_fixture('nanaimo_bar')
    assert nanaimo_bar is not None
    assert 'nanaimo_bar' == nanaimo_bar.name
    artifacts = await nanaimo_bar.gather()
    assert artifacts is not None
    artifacts.eat()
    assert nanaimo_bar.loop.is_running()


@pytest.mark.xfail
def test_assert_success() -> None:
    nanaimo.pytest.plugin.assert_success(nanaimo.Artifacts(1))



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
#                                       (@@@@%%%%%%%%%&@@&.
#                              /%&&%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%&@@(
#                              *@&%%%%%%%%%&&%%%%%%%%%%%%%%%%%%&&&%%%%%%%
#                               @   @@@(@@@@%%%%%%%%%%%%%%%%&@@&* @@@   .
#                               ,   .        .  .@@@&                   /
#                                .       .                              *
#                               @@              .                       @
#                              @&&&&&&@. .    .                     *@%&@
#                              &&&&&&&&&&&&&&&&@@        *@@############@
#                     *&/ @@ #&&&&&&&&&&&&&&&&&&&&@  ###################*
#                              @&&&&&&&&&&&&&&&&&&##################@
#                                 %@&&&&&&&&&&&&&&################@
#                                        @&&&&&&&&&&%#######&@%
#  nanaimo                                   (@&&&&####@@*
#
"""
The parsers for popular data formats used by IMUs.
"""
import codecs
import re
import typing


class YPR:
    """
    Parser for Yaw, Pitch, and Roll data. This is a line-oriented push parser.

    .. invisible-code-block: python

        from nanaimo.parsers.imu import YPR
        import pytest

    .. code-block:: python

        line0 = '#YPR=103.87,-3.60,96.24'

        # In the simplest manifestation, where you already have an ASCII decoded line,
        # you can use the parse_line method without instantiating the parser class.
        result = YPR.parse_line(line0)

    .. invisible-code-block: python

        assert result is not None
        assert result[0] == pytest.approx(103.87)
        assert result[1] == pytest.approx(-3.60)
        assert result[2] == pytest.approx(96.24)

    .. code-block:: python

        # When dealing with binary data...
        binary_data = [
            (
                '06.57,-4.74,94.76\n' +
                '#YPR=104.65,-3.73,96.25\n' +
                '#YPR=103.87,-3.60,96.24\n' +
                '#YPR=102.71,-3.45,99.21\n' +
                '#YPR=101.29,-3.39,99.14\n' +
                '\n' +
                '# This is a comment after a blank line.\n' +
                '#YPR=100.11,-3.44,99.06\n' +
                '#YPR=98.0'
            ).encode('ascii'),

            (
                '5,-2.12,100.53\n' +
                '#YPR=97.29,-2.89,99.08\n' +
                '#YPR=97.26,-3.72,97.16\n' +
                '#YPR=97.99,-3.98,96.61\n' +
                '#YPR=100.11,-4.45,94.76\n' +
                '# Note that this last record must not be handled\n' +
                '# since we cannot know if it was truncated or not\n' +
                '# because the eol character was not found yet.\n' +
                '#YPR=103.62,-5.19,9'
            ).encode('ascii')
        ]

        # ...you'll need a callback like this.
        class Handler:
            '''
            Using a functor works well if you want to accumulate
            records but you can also just use a simple function
            as a callback.
            '''
            def __init__(self):
                self.records = []

            def __call__(self, record):
                self.records.append(record)


        parser = YPR(Handler())

        # This is a fully incremental push parser so you
        # can feed it raw data from a sensor.
        for chunk in binary_data:
            parser.push_bytes(chunk)

    .. invisible-code-block: python

        handler = parser._record_callback

        assert len(handler.records) == 10
        assert parser.line_errors == 0
        assert parser.ignored_lines == 6
        assert parser.bytes_discarded == 0
        assert handler.records[5][0] == pytest.approx(98.05)
        assert handler.records[5][1] == pytest.approx(-2.12)
        assert handler.records[5][2] == pytest.approx(100.53)

    """
    ypr_pattern = re.compile(r'\s*(-?\d+(?:(?<=.)\.\d+|))\s*,?')

    default_line_ending = '\n'

    max_line_length = 254

    def __init__(self,
                 on_record: typing.Callable[[typing.Tuple[float, float, float]], None],
                 line_ending: str = default_line_ending) -> None:
        self._record_callback = on_record
        self.eol = line_ending
        self._line_buffer = ''
        self._decoder = codecs.getincrementaldecoder('ascii')('strict')
        self.reset()

    def reset(self) -> None:
        self.line_errors = 0
        self.ignored_lines = 0
        self.bytes_discarded = 0

    @classmethod
    def parse_line(cls, line: str) -> typing.Optional[typing.Tuple[float, float, float]]:
        if line.startswith('#YPR='):
            matches = cls.ypr_pattern.findall(line[5:])
            if 3 == len(matches):
                return (float(matches[0]), float(matches[1]), float(matches[2]))
            else:
                raise ValueError('Line starting with #YPR= did not conform to the expected structure.')
        return None

    def push_bytes(self, ypr_bytes: bytes) -> None:
        decoded = self._decoder.decode(ypr_bytes)
        decoded_w_buffer = self._line_buffer + decoded
        self._line_buffer = ''
        lines = decoded_w_buffer.split(self.eol)

        if not decoded.endswith(self.eol):
            if len(lines) > 0:
                self._line_buffer = lines[-1]
                lines = lines[:-1]
            else:
                self._line_buffer = decoded

            if len(self._line_buffer) > self.max_line_length:
                # ASCII == 8 bits
                self.bytes_discarded += len(self._line_buffer)
                self._line_buffer = ''

        for line in lines:
            try:
                ypr = self.parse_line(line)
                if ypr is None:
                    self.ignored_lines += 1
                else:
                    self._record_callback(ypr)
            except ValueError:
                self.line_errors += 1
